use crate::{
    gui_handler::gui::displayable::{DisplayableFillType, DisplayableType},
    prelude::*,
};
use anyhow::Result;
use assetpath::AssetPath;
use utilities::prelude::*;

use std::sync::{
    Arc,
    atomic::{AtomicBool, Ordering::SeqCst},
};

#[derive(Debug, Clone)]
pub enum FillTypeInfo {
    Image(AssetPath),
    Color(Color, DisplayableFillType),
    Element(ElementDescriptor, DisplayableFillType),
}

impl From<(Color, DisplayableFillType)> for FillTypeInfo {
    fn from((color, fill_type): (Color, DisplayableFillType)) -> Self {
        Self::Color(color, fill_type)
    }
}

impl From<Color> for FillTypeInfo {
    fn from(color: Color) -> Self {
        Self::Color(color, DisplayableFillType::Expand)
    }
}

impl From<String> for FillTypeInfo {
    fn from(s: String) -> Self {
        Self::Image(AssetPath::from(s))
    }
}

impl From<&str> for FillTypeInfo {
    fn from(s: &str) -> Self {
        Self::Image(AssetPath::from(s))
    }
}

impl From<&String> for FillTypeInfo {
    fn from(s: &String) -> Self {
        Self::Image(AssetPath::from(s.as_str()))
    }
}

impl From<AssetPath> for FillTypeInfo {
    fn from(assetpath: AssetPath) -> Self {
        Self::Image(assetpath)
    }
}

pub(crate) enum InnerFillType {
    Image(Arc<Displayable>),
    Color(Arc<Colorable>),
}

impl InnerFillType {
    pub(crate) fn new(
        gui_handler: &mut GuiHandler,
        framable: Arc<Framable>,
        fill_type_info: FillTypeInfo,
    ) -> Result<Self> {
        match fill_type_info {
            FillTypeInfo::Element(descriptor, fill_type) => Ok(InnerFillType::Image(
                Displayable::new(gui_handler, framable, descriptor, fill_type)?,
            )),
            FillTypeInfo::Image(path) => Ok(InnerFillType::Image(Displayable::new(
                gui_handler,
                framable,
                path,
                DisplayableFillType::Expand,
            )?)),
            FillTypeInfo::Color(color, fill_type) => Ok(InnerFillType::Color(Colorable::new(
                gui_handler,
                framable,
                color,
                fill_type,
            )?)),
        }
    }

    pub(crate) fn enable(&self, gui_handler: &mut GuiHandler) -> Result<()> {
        match self {
            Self::Image(displayable) => displayable.add(gui_handler)?,
            Self::Color(colorable) => colorable.add(gui_handler)?,
        }

        Ok(())
    }

    pub(crate) fn disable(&self, gui_handler: &mut GuiHandler) -> Result<()> {
        match self {
            Self::Image(displayable) => displayable.delete(gui_handler)?,
            Self::Color(colorable) => colorable.delete(gui_handler)?,
        }

        Ok(())
    }

    pub(crate) fn update_frame(&self, gui_handler: &mut GuiHandler) -> Result<()> {
        match self {
            Self::Image(displayable) => displayable.update_frame(gui_handler)?,
            Self::Color(colorable) => colorable.update_frame(gui_handler)?,
        }

        Ok(())
    }
}

impl Drop for InnerFillType {
    fn drop(&mut self) {
        match self {
            Self::Image(displayable) => displayable.clear_callback(),
            Self::Color(colorable) => colorable.clear_callback(),
        }
    }
}

pub(crate) struct FillType {
    pub(crate) inner: InnerFillType,
    visible: AtomicBool,
}

impl FillType {
    pub(crate) fn new(
        gui_handler: &mut GuiHandler,
        framable: Arc<Framable>,
        fill_type_info: FillTypeInfo,
    ) -> Result<Self> {
        let me = Self {
            inner: InnerFillType::new(gui_handler, framable.clone(), fill_type_info)?,
            visible: AtomicBool::new(false),
        };

        me.set_ui_layer(framable.ui_layer());

        Ok(me)
    }

    pub(crate) fn enable(&self, gui_handler: &mut GuiHandler) -> Result<()> {
        if !self.visible.load(SeqCst) {
            self.visible.store(true, SeqCst);
            self.inner.enable(gui_handler)?;
        }

        Ok(())
    }

    pub(crate) fn disable(&self, gui_handler: &mut GuiHandler) -> Result<()> {
        if self.visible.load(SeqCst) {
            self.visible.store(false, SeqCst);
            self.inner.disable(gui_handler)?;
        }

        Ok(())
    }

    pub(crate) fn update_frame(&self, gui_handler: &mut GuiHandler) -> Result<()> {
        self.inner.update_frame(gui_handler)
    }

    pub(crate) fn set_ui_layer(&self, layer: i32) {
        match &self.inner {
            InnerFillType::Image(displayable) => displayable.set_ui_layer(layer),
            InnerFillType::Color(colorable) => colorable.set_ui_layer(layer),
        }
    }
}

impl PartialEq<FillTypeInfo> for FillType {
    fn eq(&self, info: &FillTypeInfo) -> bool {
        match (&self.inner, info) {
            (InnerFillType::Image(inner_image), FillTypeInfo::Image(info_image)) => {
                match inner_image.inner_type() {
                    DisplayableType::Path(path) => path == *info_image,
                    DisplayableType::Descriptor(_) => false,
                }
            }
            (InnerFillType::Image(inner_image), FillTypeInfo::Element(info_descriptor, _)) => {
                match inner_image.inner_type() {
                    DisplayableType::Path(_) => false,
                    DisplayableType::Descriptor(descriptor) => descriptor == *info_descriptor,
                }
            }
            (InnerFillType::Color(inner_color), FillTypeInfo::Color(info_color, _)) => {
                inner_color.color() == *info_color
            }

            _ => false,
        }
    }
}
