//! `Clickable` is a property to click an item per mouse

use anyhow::Result;

use super::executable::Executable;
use super::framable::Framable;
use crate::prelude::*;

#[cfg(feature = "audio")]
use super::audible::Audible;

use std::sync::{
    Arc, RwLock,
    atomic::{AtomicBool, AtomicI32, Ordering::SeqCst},
};

/// `Clickable` gives the ability to execute a closure when clicked
pub struct Clickable {
    framable: Arc<Framable>,

    #[cfg(feature = "audio")]
    audible: RwLock<Option<Arc<Audible>>>,

    clicked: AtomicBool,
    ui_layer: AtomicI32,

    executable: Arc<Executable<()>>,
    clicked_changed_callback:
        RwLock<Option<Box<dyn Fn(&mut GuiHandler) -> Result<()> + Send + Sync>>>,
}

impl Clickable {
    /// Factory method for `Clickable`, returns `Arc<Clickable>`
    ///
    /// # Arguments
    ///
    /// * `framable` is a `Arc<Framable>` instance
    /// * `executable` is a `Arc<Executable>` instance
    pub fn new(framable: Arc<Framable>, executable: Arc<Executable<()>>) -> Arc<Self> {
        Arc::new(Clickable {
            framable,

            #[cfg(feature = "audio")]
            audible: RwLock::new(None),

            clicked: AtomicBool::new(false),
            ui_layer: AtomicI32::new(0),

            executable,
            clicked_changed_callback: RwLock::new(None),
        })
    }

    /// Add method
    ///
    /// # Arguments
    ///
    /// * `clickable` is a `&Arc<Clickable>` instance that is going to be added
    pub fn add(self: &Arc<Self>, gui_handler: &mut GuiHandler) -> Result<()> {
        gui_handler.add_clickable(self.ui_layer.load(SeqCst), self.clone())
    }

    /// Delete method, has to be explicitly called, otherwise it will remain in memory
    ///
    /// # Arguments
    ///
    /// * `clickable` is a `&Arc<Clickable>` instance that is going to be deleted
    pub fn delete(self: &Arc<Self>, gui_handler: &mut GuiHandler) -> Result<()> {
        gui_handler.delete_clickable(self.ui_layer.load(SeqCst), self)?;
        self.set_clicked(gui_handler, false)?;

        Ok(())
    }

    pub fn set_ui_layer(&self, ui_layer: i32) {
        self.ui_layer.store(ui_layer, SeqCst);
    }

    pub fn is_pressed(&self, x: i32, y: i32) -> bool {
        (x > self.framable.left())
            && (x < self.framable.right())
            && (y > self.framable.top())
            && (y < self.framable.bottom())
    }

    /// Callback when the clicked status of `self` has changed
    ///
    /// # Arguments
    ///
    /// * `clicked_changed_callback` is a `Option<Box<Callback>>`
    pub fn set_clicked_changed_callback(
        &self,
        clicked_changed_callback: Option<Box<dyn Fn(&mut GuiHandler) -> Result<()> + Send + Sync>>,
    ) {
        *self.clicked_changed_callback.write().unwrap() = clicked_changed_callback;
    }

    #[cfg(feature = "audio")]
    pub fn set_audible(&self, audible: Option<Arc<Audible>>) {
        *self.audible.write().unwrap() = audible;
    }

    pub fn set_clicked(&self, gui_handler: &mut GuiHandler, clicked: bool) -> Result<bool> {
        if self.clicked() != clicked {
            self.clicked.store(clicked, SeqCst);

            if let Some(clicked_changed_callback) =
                self.clicked_changed_callback.read().unwrap().as_ref()
            {
                clicked_changed_callback(gui_handler)?;
            }

            if self.clicked() {
                #[cfg(feature = "audio")]
                {
                    if let Some(audible) = self.audible.read().unwrap().as_ref() {
                        audible.play()?;
                    }
                }

                self.executable.execute(gui_handler, ())?;
            }

            Ok(true)
        } else {
            Ok(false)
        }
    }

    /// Get current `clicked` status, returns `bool`
    pub fn clicked(&self) -> bool {
        self.clicked.load(SeqCst)
    }
}
